/********************************************************************/
/**                                                                **/
/**   exapmle_4.c                                                  **/
/**                                                                **/
/**   A universal generator for multivariate T-concave             **/
/**   distributions                                                **/
/**                                                                **/
/**   generate normal distributed variates                         **/
/**                                                                **/
/**   domain:  IR^n                                                **/
/**   mode:    origin                                              **/
/**   density: f(x_1,...x_N) = exp(-\sum_{i=1}^N i*x_i^2)          **/
/**                                                                **/
/**                                                                **/
/**   authors: Josef Leydold                                       **/
/**   last modified: Mon Mar  9 13:38:00 MET 1998                  **/
/**                                                                **/
/********************************************************************/              

/* the header file */
#include "tdrmv.h"

/*------------------------------------------------------------------*/
/* Compiling options in config.h                                    */
/*                                                                  */
/*    N          2, ..., 14                                         */
/*    MODE       0 or 1                                             */
/*    RECTANGLE  0                                                  */
/*    GAMMA      1 or 2                                             */
/*                                                                  */
/*------------------------------------------------------------------*/

#define SAMPLE_SIZE   1000     /* sample size                       */

/*------------------------------------------------------------------*/

/* print random point */
static void print_rpoint(double *v, int n);

/*------------------------------------------------------------------*/

/*******************************************************************

  Main

  ******************************************************************/

int main()
{
  C_ROOT *hat;         /* pointer to hat */
  C_CONSTR *hc;        /* parameters for construting hat functions */
  double rpoint[N];    /* array for random point */
  int i;

  /* density */
  double initialize_normal();      /* initialize density function */
  double normal();                 /* density function */
  void grad_normal();              /* gradient of density */
  
  /*--------------------------------------------------------------*/

  /* default parameters for constructing hat function */
  hc = get_constr_default();

  /*--------------------------------------------------------------*/

  /* volume below density (default is 0. == unknown) */
  hc->volume_density = initialize_normal();

  /*--------------------------------------------------------------*/

  /* hat */
  hat = get_hat(hc,normal,grad_normal);

  /*--------------------------------------------------------------*/

  /* sample */
  for( i=0; i<SAMPLE_SIZE; i++ ) {
    sample_point(hat,rpoint);
    print_rpoint(rpoint,N);
  }

  /*--------------------------------------------------------------*/

#if DEBUG > 0
  db_close(hat);
#endif
  
  exit(0);
} /* end of main() */

/*-----------------------------------------------------------------*/
static void print_rpoint(double *v, int n)
/*
  print random point
*/
{
  int i;

  printf("( %f",v[0]);
  for( i=1; i<n; i++ )
    printf(", %f",v[i]);
  printf(" )\n");

} /* end of db_print_rpoint() */

/*-----------------------------------------------------------------*/

/*******************************************************************

  normal distribution

  ******************************************************************/

/*-----------------------------------------------------------------*/

double initialize_normal(void)
/*
  returns volume below density function
*/
{
  int i;
  double volume_density;   /* volume below density */
  
  /* compute volume below density */
  volume_density = 1.;
  for( i=0; i<N; i++ )
    volume_density *= PI / (i+1.);
  return( sqrt(volume_density) );

} /* end of initialize_normal() */

/*-----------------------------------------------------------------*/

double normal(double *coord)
/*
  normal density function
*/
{
  int i;             /* aux counter */
  double f=0;       /* density function */

  /* gaussian density */
  for( i=0; i<N; i++ )
    f += (i+1) * coord[i] * coord[i];

  return( exp(-f) );
} /* end of normal() */

/*-----------------------------------------------------------------*/

void grad_normal(double *coord, double *grad)
/*
  gradient of normal density function
*/
{
  int i;              /* aux counter */
  double f=0;             /* density at point */

  /* d(e^{-\sum j*x_j^2})/d x_i = -2*i*x_i*e^{-\sum j*x_j^2} */

  for( i=0; i<N; i++ )
    f += (i+1) * coord[i] * coord[i];
  f = exp( -f );

  for( i=0; i<N; i++ )
    grad[i] = (-2) * (i+1) * coord[i] * f;

} /* end of grad_normal() */

/*-----------------------------------------------------------------*/

